/*
 (c) Copyright, Real-Time Innovations, $Date: 2006/09/16 06:56:14 $.
 All rights reserved.
 No duplications, whole or partial, manual or electronic, may be made
 without express written permission.  Any such copies, or
 revisions thereof, must display this notice unaltered.
 This code contains trade secrets of Real-Time Innovations, Inc.
*/



// ---------------------------------------------------------------------------
// -- BaseThroughputListener
// ---------------------------------------------------------------------------
class BaseThroughputListener : public DDSDataReaderListener {

  protected:
    DDS_UnsignedLong _sequence_number;
    int _matching_writer;
    int _test_verbosity;

    /* Used to estimate the size of each batch in case we miss a sample.
       When a batch is reported lost, the samples lost are estimated
       by multiplying the gap in sequence numberd by this value. */
    int _size_of_last_batch;
    /* All these counters count the contained Throughput packets, not the 
     * Batched ones */
    int _packets_received;
    int _packets_lost;
    int _samples_lost;
    int _samples_rejected;


  public:
    ~BaseThroughputListener() {};

    BaseThroughputListener() {
        reset();
        _test_verbosity = 0;
    }

    virtual void on_requested_deadline_missed(
        DDSDataReader* /*reader*/,
        const DDS_RequestedDeadlineMissedStatus& /*status*/) {}

    virtual void on_requested_incompatible_qos(
        DDSDataReader* /*reader*/,
        const DDS_RequestedIncompatibleQosStatus& /*status*/) {}

    virtual void on_sample_rejected(
        DDSDataReader* reader, const DDS_SampleRejectedStatus& status) {
        _samples_rejected += status.total_count_change*_size_of_last_batch;
    }

    virtual void on_liveliness_changed(
        DDSDataReader* /*reader*/,
        const DDS_LivelinessChangedStatus& /*status*/) {}

    virtual void on_sample_lost(
        DDSDataReader* reader, const DDS_SampleLostStatus& status) {
            _samples_lost += status.total_count_change*_size_of_last_batch;
    }

    virtual void on_subscription_matched(
        DDSDataReader* reader, const DDS_SubscriptionMatchedStatus& status) {
       _matching_writer = status.current_count;
    }

    void set_verbosity(int verbosity) { _test_verbosity = verbosity; }
    int get_packets_received() { return _packets_received; }
    int get_packets_lost() { return _packets_lost; }
    int get_samples_lost() { return _samples_lost; }
    int get_samples_rejected() { return _samples_rejected; }
    int get_matching_writers() { return _matching_writer; }

    void resetCounts() {
       _samples_rejected = 0;
       _samples_lost = 0;
       _sequence_number = 0;
       _packets_received = 0;
       _packets_lost = 0;
    }

    void reset() {
        resetCounts();
        _matching_writer = 0;
        _size_of_last_batch = 1;
    }
};




// ---------------------------------------------------------------------------
// -- BatchedThroughputListener
// ---------------------------------------------------------------------------
class BatchedThroughputListener : public BaseThroughputListener {
    virtual void BatchedThroughputListener::on_data_available(DDSDataReader* reader);
};

void BatchedThroughputListener::on_data_available(DDSDataReader* reader) {
    /* Increment number of packets received */
    int i, j;
    ThroughputSequenceDataReader* data_reader;
    ThroughputSequenceSeq data_seq;
    DDS_SampleInfoSeq info_seq;
    DDS_ReturnCode_t return_code;

    data_reader = ThroughputSequenceDataReader::narrow(reader);

    return_code = data_reader->take(data_seq, info_seq, DDS_LENGTH_UNLIMITED,
                                    DDS_NOT_READ_SAMPLE_STATE,
                                    DDS_ANY_VIEW_STATE,
                                    DDS_ALIVE_INSTANCE_STATE);
    if (return_code != DDS_RETCODE_OK) {
        if (return_code != DDS_RETCODE_NO_DATA) {
            AppLog_exception("Error in reading data\n");
        }
        return;
    }

    // with each data sequence
    for(i = 0; i < data_seq.length(); ++i) {
	if ( info_seq[i].valid_data ) {

	    // with each data sample
	    for ( j = 0; j < data_seq[i].elem_seq.length(); ++j) {

		// is the sequence number what we expect?
		if (data_seq[i].elem_seq[j].sequence_number > _sequence_number) {
		    
                    if (_sequence_number != 0) {  // if this is not the first expected data
			if (_test_verbosity >= 1) {
			    printf("dataSeq[%d].elem_seq[%d].sequence_number = %d, Expected seq = %d\n",
				i, j, 
				data_seq[i].elem_seq[j].sequence_number, 
				_sequence_number);
			}
			_packets_lost += (data_seq[i].elem_seq[j].sequence_number - _sequence_number);
                    }

		    // Reset sequence number
		    _sequence_number = data_seq[i].elem_seq[j].sequence_number;
		}


		++_packets_received;
		++_sequence_number;  // Increment expected sequence number
	    }
            _size_of_last_batch  = data_seq[i].elem_seq.length();
	} 
    }

    data_reader->return_loan(data_seq, info_seq);
};




// ---------------------------------------------------------------------------
// -- ThroughputListener
// ---------------------------------------------------------------------------
class ThroughputListener : public BaseThroughputListener {
public:
    virtual void on_data_available(DDSDataReader* reader);
 
};

void ThroughputListener::on_data_available(DDSDataReader* reader) {
    /* Increment number of packets received */
    int i;
    ThroughputDataReader* data_reader;
    ThroughputSeq data_seq;
    DDS_SampleInfoSeq info_seq;
    DDS_ReturnCode_t return_code;

    data_reader = ThroughputDataReader::narrow(reader);

    return_code = data_reader->take(data_seq, info_seq, DDS_LENGTH_UNLIMITED,
                                    DDS_NOT_READ_SAMPLE_STATE,
                                    DDS_ANY_VIEW_STATE,
                                    DDS_ALIVE_INSTANCE_STATE);
    if (return_code != DDS_RETCODE_OK) {
        if (return_code != DDS_RETCODE_NO_DATA) {
            AppLog_exception("Error in reading data\n");
        }
        return;
    }
    for (i = 0; i < data_seq.length(); ++i) {
        if (info_seq[i].valid_data) {
            if (data_seq[i].sequence_number != _sequence_number) {
                if (_sequence_number != 0) {
                    if (_test_verbosity >= TEST_VERBOSITY_MESSAGES) {
                        AppLog_warn("data_seq[%d].sequence_number = %d, "
                                    "Expected sequence_number = %d\n",
                                    i, data_seq[i].sequence_number,
                                    _sequence_number);
                    }
                    _packets_lost += (data_seq[i].sequence_number -
                        _sequence_number);
                }
                /* Reset sequence number */
                _sequence_number = data_seq[i].sequence_number;
            }
            ++_packets_received;
            ++_sequence_number;
        }
    }
    data_reader->return_loan(data_seq, info_seq);
}


